# encoding: utf-8
'''Reserved tokens.'''

RESERVED = (
    "abstract",
    "alias",
    "and",
    "as",
    "begin",
    "break",
    "byte",
    "case",
    "catch",
    "class",
    "clone",
    "const",
    "continue",
    "declare",
    "def",
    "default",
    "defined",
    "descriptor",
    "descriptors",
    "do",
    "elif",
    "else",
    "elseif",
    "end",
    "enddeclare",
    "endfor",
    "endforeach",
    "endif",
    "endswitch",
    "endwhile",
    "ensure",
    "extends",
    "false",
    "final",
    "for",
    "foreach",
    "function",
    "global",
    "goto",
    "if",
    "implements",
    "in",
    "instanceof",
    "int",
    "long",
    "module",
    "namespace",
    "new",
    "next",
    "nil",
    "not",
    "or",
    "private",
    "protected",
    "public",
    "redo",
    "rescue",
    "retry",
    "return",
    "self",
    "short",
    "static",
    "super",
    "switch",
    "then",
    "throw",
    "true",
    "try",
    "undef",
    "unless",
    "until",
    "use",
    "var",
    "when",
    "while",
    "xor",
    "yield"
)
